 (**      HasCart.Pas by Alexander Myasnikov                        **)
 (**      HashCart Hash function  based on Cartman block cipher     **)
 (**      Freeware for any use,  non-patented, opensource cipher    **)
 (**      WEB:       www.darksoftware.narod.ru                      **)
 (**      Cartman block cipher project information                  **)
 (**      WEB:       www.alexanderwdark.narod.ru                    **)


unit hashcart2ddp2;


interface

uses Classes;

type
  THashCart = class(TObject)

  public
    procedure Update (Data: Pointer; Len: integer; FirstBlock: boolean = False;
      LastBlock: boolean = False);
    procedure Final (Digest: pointer);
    procedure CalcStream (const Stream: TStream);
    procedure CalcFile (const FileName: string);
    procedure CalcString (const S: string);
    procedure Burn;

    constructor Create (Mix: integer = 5);
  private
    Keystream:  array [0..3] of longword;
    Mix_cycles: integer;
  end;

var
  HCart: THashCart = nil;

implementation

uses hccartman2ddp2, Windows;

const
  H_Data: array[0..15] of longword =
    ($452821E6, $38D01377, $BE5466CF, $34E90C6C, $C0AC29B7, $C97C50DD,
    $3F84D5B5, $B5470917,
    $9216D5D9, $8979FB1B, $D1310BA6, $98DFB5AC, $2FFD72DB, $D01ADFB7,
    $B8E1AFED, $6A267E96);

const
  SBOX: array[0..4, 0..255] of byte = (


    ($B1, $CE, $C3, $95, $5A, $AD, $E7, $02, $4D, $44, $FB, $91, $0C, $87,
    $A1, $50, $CB, $67, $54, $DD, $46, $8F, $E1, $4E, $F0, $FD, $FC, $EB,
    $F9, $C4, $1A, $6E, $5E, $F5, $CC, $8D, $1C, $56, $43, $FE, $07, $61,
    $F8, $75, $59, $FF, $03, $22, $8A, $D1, $13, $EE, $88, $00, $0E, $34,
    $15, $80, $94, $E3, $ED, $B5, $53, $23, $4B, $47, $17, $A7, $90, $35,
    $AB, $D8, $B8, $DF, $4F, $57, $9A, $92, $DB, $1B, $3C, $C8, $99, $04,
    $8E, $E0, $D7, $7D, $85, $BB, $40, $2C, $3A, $45, $F1, $42, $65, $20,
    $41, $18, $72, $25, $93, $70, $36, $05, $F2, $0B, $A3, $79, $EC, $08,
    $27, $31, $32, $B6, $7C, $B0, $0A, $73, $5B, $7B, $B7, $81, $D2, $0D,
    $6A, $26, $9E, $58, $9C, $83, $74, $B3, $AC, $30, $7A, $69, $77, $0F,
    $AE, $21, $DE, $D0, $2E, $97, $10, $A4, $98, $A8, $D4, $68, $2D, $62,
    $29, $6D, $16, $49, $76, $C7, $E8, $C1, $96, $37, $E5, $CA, $F4, $E9,
    $63, $12, $C2, $A6, $14, $BC, $D3, $28, $AF, $2F, $E6, $24, $52, $C6,
    $A0, $09, $BD, $8C, $CF, $5D, $11, $5F, $01, $C5, $9F, $3D, $A2, $9B,
    $C9, $3B, $BE, $51, $19, $1F, $3F, $5C, $B2, $EF, $4A, $CD, $BF, $BA,
    $6F, $64, $D9, $F3, $3E, $B4, $AA, $DC, $D5, $06, $C0, $7E, $F6, $66,
    $6C, $84, $71, $38, $B9, $1D, $7F, $9D, $48, $8B, $2A, $DA, $A5, $33,
    $82, $39, $D6, $78, $86, $FA, $E4, $2B, $A9, $1E, $89, $60, $6B, $EA,
    $55, $4C, $F7, $E2),

    ($35, $BE, $07, $2E, $53, $69, $DB, $28, $6F, $B7, $76, $6B, $0C, $7D,
    $36, $8B, $92, $BC, $A9, $32, $AC, $38, $9C, $42, $63, $C8, $1E, $4F,
    $24, $E5, $F7, $C9, $61, $8D, $2F, $3F, $B3, $65, $7F, $70, $AF, $9A,
    $EA, $F5, $5B, $98, $90, $B1, $87, $71, $72, $ED, $37, $45, $68, $A3,
    $E3, $EF, $5C, $C5, $50, $C1, $D6, $CA, $5A, $62, $5F, $26, $09, $5D,
    $14, $41, $E8, $9D, $CE, $40, $FD, $08, $17, $4A, $0F, $C7, $B4, $3E,
    $12, $FC, $25, $4B, $81, $2C, $04, $78, $CB, $BB, $20, $BD, $F9, $29,
    $99, $A8, $D3, $60, $DF, $11, $97, $89, $7E, $FA, $E0, $9B, $1F, $D2,
    $67, $E2, $64, $77, $84, $2B, $9E, $8A, $F1, $6D, $88, $79, $74, $57,
    $DD, $E6, $39, $7B, $EE, $83, $E1, $58, $F2, $0D, $34, $F8, $30, $E9,
    $B9, $23, $54, $15, $44, $0B, $4D, $66, $3A, $03, $A2, $91, $94, $52,
    $4C, $C3, $82, $E7, $80, $C0, $B6, $0E, $C2, $6C, $93, $EC, $AB, $43,
    $95, $F6, $D8, $46, $86, $05, $8C, $B0, $75, $00, $CC, $85, $D7, $3D,
    $73, $7A, $48, $E4, $D1, $59, $AD, $B8, $C6, $D0, $DC, $A1, $AA, $02,
    $1D, $BF, $B5, $9F, $51, $C4, $A5, $10, $22, $CF, $01, $BA, $8F, $31,
    $7C, $AE, $96, $DA, $F0, $56, $47, $D4, $EB, $4E, $D9, $13, $8E, $49,
    $55, $16, $FF, $3B, $F4, $A4, $B2, $06, $A0, $A7, $FB, $1B, $6E, $3C,
    $33, $CD, $18, $5E, $6A, $D5, $A6, $21, $DE, $FE, $2A, $1C, $F3, $0A,
    $1A, $19, $27, $2D),

    ($A3, $D7, $09, $83, $F8, $48, $F6, $F4, $B3, $21, $15, $78, $99, $B1,
    $AF, $F9, $E7, $2D, $4D, $8A, $CE, $4C, $CA, $2E, $52, $95, $D9, $1E,
    $4E, $38, $44, $28, $0A, $DF, $02, $A0, $17, $F1, $60, $68, $12, $B7,
    $7A, $C3, $E9, $FA, $3D, $53, $96, $84, $6B, $BA, $F2, $63, $9A, $19,
    $7C, $AE, $E5, $F5, $F7, $16, $6A, $A2, $39, $B6, $7B, $0F, $C1, $93,
    $81, $1B, $EE, $B4, $1A, $EA, $D0, $91, $2F, $B8, $55, $B9, $DA, $85,
    $3F, $41, $BF, $E0, $5A, $58, $80, $5F, $66, $0B, $D8, $90, $35, $D5,
    $C0, $A7, $33, $06, $65, $69, $45, $00, $94, $56, $6D, $98, $9B, $76,
    $97, $FC, $B2, $C2, $B0, $FE, $DB, $20, $E1, $EB, $D6, $E4, $DD, $47,
    $4A, $1D, $42, $ED, $9E, $6E, $49, $3C, $CD, $43, $27, $D2, $07, $D4,
    $DE, $C7, $67, $18, $89, $CB, $30, $1F, $8D, $C6, $8F, $AA, $C8, $74,
    $DC, $C9, $5D, $5C, $31, $A4, $70, $88, $61, $2C, $9F, $0D, $2B, $87,
    $50, $82, $54, $64, $26, $7D, $03, $40, $34, $4B, $1C, $73, $D1, $C4,
    $FD, $3B, $CC, $FB, $7F, $AB, $E6, $3E, $5B, $A5, $AD, $04, $23, $9C,
    $14, $51, $22, $F0, $29, $79, $71, $7E, $FF, $8C, $0E, $E2, $0C, $EF,
    $BC, $72, $75, $6F, $37, $A1, $EC, $D3, $8E, $62, $8B, $86, $10, $E8,
    $08, $77, $11, $BE, $92, $4F, $24, $C5, $32, $36, $9D, $CF, $F3, $A6,
    $BB, $AC, $5E, $6C, $A9, $13, $57, $25, $B5, $E3, $BD, $A8, $3A, $01,
    $05, $59, $2A, $46),

    ($20, $89, $EF, $BC, $66, $7D, $DD, $48, $D4, $44, $51, $25, $56, $ED,
    $93, $95, $46, $E5, $11, $7C, $73, $CF, $21, $14, $7A, $8F, $19, $D7,
    $33, $B7, $8A, $8E, $92, $D3, $6E, $AD, $01, $E4, $BD, $0E, $67, $4E,
    $A2, $24, $FD, $A7, $74, $FF, $9E, $2D, $B9, $32, $62, $A8, $FA, $EB,
    $36, $8D, $C3, $F7, $F0, $3F, $94, $02, $E0, $A9, $D6, $B4, $3E, $16,
    $75, $6C, $13, $AC, $A1, $9F, $A0, $2F, $2B, $AB, $C2, $AF, $B2, $38,
    $C4, $70, $17, $DC, $59, $15, $A4, $82, $9D, $08, $55, $FB, $D8, $2C,
    $5E, $B3, $E2, $26, $5A, $77, $28, $CA, $22, $CE, $23, $45, $E7, $F6,
    $1D, $6D, $4A, $47, $B0, $06, $3C, $91, $41, $0D, $4D, $97, $0C, $7F,
    $5F, $C7, $39, $65, $05, $E8, $96, $D2, $81, $18, $B5, $0A, $79, $BB,
    $30, $C1, $8B, $FC, $DB, $40, $58, $E9, $60, $80, $50, $35, $BF, $90,
    $DA, $0B, $6A, $84, $9B, $68, $5B, $88, $1F, $2A, $F3, $42, $7E, $87,
    $1E, $1A, $57, $BA, $B6, $9A, $F2, $7B, $52, $A6, $D0, $27, $98, $BE,
    $71, $CD, $72, $69, $E1, $54, $49, $A3, $63, $6F, $CC, $3D, $C8, $D9,
    $AA, $0F, $C6, $1C, $C0, $FE, $86, $EA, $DE, $07, $EC, $F8, $C9, $29,
    $B1, $9C, $5C, $83, $43, $F9, $F5, $B8, $CB, $09, $F1, $00, $1B, $2E,
    $85, $AE, $4B, $12, $5D, $D1, $64, $78, $4C, $D5, $10, $53, $04, $6B,
    $8C, $34, $3A, $37, $03, $F4, $61, $C5, $EE, $E3, $76, $31, $4F, $E6,
    $DF, $A5, $99, $3B),

    ($63, $7C, $77, $7B, $F2, $6B, $6F, $C5, $30, $01, $67, $2B, $FE, $D7,
    $AB, $76, $CA, $82, $C9, $7D, $FA, $59, $47, $F0, $AD, $D4, $A2, $AF,
    $9C, $A4, $72, $C0, $B7, $FD, $93, $26, $36, $3F, $F7, $CC, $34, $A5,
    $E5, $F1, $71, $D8, $31, $15, $04, $C7, $23, $C3, $18, $96, $05, $9A,
    $07, $12, $80, $E2, $EB, $27, $B2, $75, $09, $83, $2C, $1A, $1B, $6E,
    $5A, $A0, $52, $3B, $D6, $B3, $29, $E3, $2F, $84, $53, $D1, $00, $ED,
    $20, $FC, $B1, $5B, $6A, $CB, $BE, $39, $4A, $4C, $58, $CF, $D0, $EF,
    $AA, $FB, $43, $4D, $33, $85, $45, $F9, $02, $7F, $50, $3C, $9F, $A8,
    $51, $A3, $40, $8F, $92, $9D, $38, $F5, $BC, $B6, $DA, $21, $10, $FF,
    $F3, $D2, $CD, $0C, $13, $EC, $5F, $97, $44, $17, $C4, $A7, $7E, $3D,
    $64, $5D, $19, $73, $60, $81, $4F, $DC, $22, $2A, $90, $88, $46, $EE,
    $B8, $14, $DE, $5E, $0B, $DB, $E0, $32, $3A, $0A, $49, $06, $24, $5C,
    $C2, $D3, $AC, $62, $91, $95, $E4, $79, $E7, $C8, $37, $6D, $8D, $D5,
    $4E, $A9, $6C, $56, $F4, $EA, $65, $7A, $AE, $08, $BA, $78, $25, $2E,
    $1C, $A6, $B4, $C6, $E8, $DD, $74, $1F, $4B, $BD, $8B, $8A, $70, $3E,
    $B5, $66, $48, $03, $F6, $0E, $61, $35, $57, $B9, $86, $C1, $1D, $9E,
    $E1, $F8, $98, $11, $69, $D9, $8E, $94, $9B, $1E, $87, $E9, $CE, $55,
    $28, $DF, $8C, $A1, $89, $0D, $BF, $E6, $42, $68, $41, $99, $2D, $0F,
    $B0, $54, $BB, $16));


type
  TLongWordArray = array [0.. 15] of longword;

type
  PLongWordArray = ^TLongWordArray;

type
  TByteArray = array [0.. 511] of byte;

type
  PByteArray = ^TByteArray;


constructor THashCart.Create (Mix: integer = 5);
begin
  Mix_cycles := Mix;
end;

procedure THashCart.burn;
begin
  FillChar(Keystream, 16, 0);
end;

procedure THashCart.update (Data: Pointer; Len: integer;
  FirstBlock: boolean = False; LastBlock: boolean = False);
var
  InitArray: array [0..15] of longword;
  IV: array[0..3] of longword;
var
  i, n, max: integer;

begin

  if FirstBlock then
    begin

    FillChar(KeyStream, 16, 0);

    for I := 0 to 15 do
      begin
      InitArray[i] := H_Data[i];
      end;

    if Len >= 16 then
      max := 16
    else
      max := Len;


    for I := 0 to max - 1 do
      begin

      PByteArray(@InitArray)^[i] := PByteArray(Data)^[i];

      end;


    hccartman_setkey(@InitArray[0]);
    Move(InitArray[12], KeyStream, 16);

    for I := 0 to 15 do
      PByteArray(@Keystream)^[i] := SBOX[0, PByteArray(@Keystream)^[i]];


    for I := 0 to mix_cycles - 1 do
      begin
      hccartman_decrypt(@KeyStream);

      for n := 0 to 15 do
        PByteArray(@Keystream)^[i] := SBOX[1, PByteArray(@Keystream)^[i]];

      end;

    end;

  Move(KeyStream, Iv, 16);

  if Len >= 16 then
    begin

    repeat
      Move(Data^, KeyStream, 16);
      for I := 0 to 3 do
        KeyStream[i] := KeyStream[i] xor Iv[i];

      for I := 0 to 15 do
        PByteArray(@Keystream)^[i] := SBOX[2, PByteArray(@Keystream)^[i]];


      Dec(len, 16);
      longword(Data) := longword(Data) + 16;
      hccartman_crypt(@KeyStream);
      Move(KeyStream, Iv, 16);
    until len < 16;

    end;

  if len > 0 then
    begin
    Move(Data^, KeyStream, len);

    for I := 0 to 15 do
      PByteArray(@Keystream)^[i] := SBOX[3, PByteArray(@Keystream)^[i]];

    for I := 0 to 3 do
      KeyStream[i] := KeyStream[i] xor Iv[i];
    hccartman_crypt(@KeyStream);
    end;

  if LastBlock then
    begin

    for I := 0 to 15 do
      begin
      PByteArray(@Keystream)^[i] := SBOX[4, PByteArray(@Keystream)^[i]];
      end;

    for n := 0 to mix_cycles - 1 do
      hccartman_crypt(@KeyStream);

    end;

end;


procedure THashCart.CalcStream (const Stream: TStream);
const
  maxBufSize = 8192;
var
  Buf: Pointer;
  BufSize, Left: integer;
  First: boolean;

begin

  Buf := nil;
    try
    Buf  := AllocMem(maxBufSize);
    Stream.Position := 0;
    Left := Stream.Size;
    First := True;

    repeat

      if Left > maxBufSize then
        BufSize := maxBufSize
      else
        BufSize := Left;
      BufSize := Stream.Read(Buf^, BufSize);
      if BufSize <= 0 then
        Break;
      Update(Buf, BufSize, First, Bufsize < maxbufsize);
      First := False;
      Dec(Left, BufSize);

    until BufSize <= 0;
    finally
    FreeMem(Buf);
    end;
end;


procedure THashCart.CalcFile (const FileName: string);
var
  S: TFileStream;
begin
  S := TFileStream.Create(FileName, 0);
    try
    CalcStream(S);
    finally
    S.Free;
    end;
end;


procedure THashCart.Final (Digest: Pointer);
begin
  Move(keystream, Digest^, 16);
end;


procedure THashCart.CalcString (const S: string);
begin
  Update(PChar(s), length(s), True);
end;

initialization
  HCart := THashCart.Create;

finalization
  HCart.Free;

end.
